;; -*- geiser-scheme-implementation: guile -*-

;; tower.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; Game configuration parameters

(define-module (td tower)
  #:use-module (td tick)
  #:use-module (td hooks)
  #:use-module (td util)
  #:use-module (td euclidean)
  #:use-module (td gfx)
  #:use-module (td picture)
  #:use-module (td fset)
  #:use-module (td enemy)
  #:use-module (td credits)
  #:use-module (aa-tree)
  #:use-module (chickadee)
  #:use-module (oop goops)
  #:export (<basic-turret>
            <basic-turret-l2>
            <basic-turret-l3>
            delete-tower-from-towers!
            <delay-turret>
            <delay-turret-l2>
            <delay-turret-l3>
            get-from-tower-selection-grid
            get-uid-from-tower-uid-grid
            place!
            tower-selection
            tower-selection-grid
            towers
            set-tower-in-uid-grid!
            <resource-tower>
            <resource-tower-l2>
            <resource-tower-l3>
            <tower>))

;; Global data structures

(define towers #nil)
(define tower-procedures #nil)
(define tower-selection #:basic-turret)
(define tower-selection-grid #nil)
(define tower-uid-grid #nil)

(define (add-tower! new-tower-id new-tower)
  (set! towers (aa-insert towers
                          new-tower-id
                          new-tower)))

(define (add-tower-procedure! proc)
  (set! tower-procedures (cons proc tower-procedures)))

(define (get-from-tower-selection-grid grid-coords)
  (catch 'key-not-found
    (lambda ()
      (aa-search tower-selection-grid grid-coords))
    (lambda (_ _') (throw 'empty-block))))

(define (get-uid-from-tower-uid-grid grid-coords)
  (catch 'key-not-found
    (lambda ()
      (aa-search tower-uid-grid grid-coords))
    (lambda (_ _') (throw 'empty-block))))

(define (set-tower-in-uid-grid! grid-coords tower-uid)
  (set! tower-uid-grid
    (aa-insert tower-uid-grid
               grid-coords
               tower-uid)))

(define (delete-tower-from-towers! uid)
  (set! towers
    (aa-delete towers uid)))

(add-hook! new-game-hook
           (lambda ()
             (set! towers #nil)
             (set! tower-procedures #nil)
             (set! tower-selection #:basic-turret)))

(define-class <tower> (<euclidean>))

(define-method (place! (t <tower>))
  (let ((new-tower-id (get-uid)))
    (add-tower! new-tower-id t)
    new-tower-id))

(define (step-reloading-time t proc)
  (if (<= (.reloading-time-left t) period)
      (begin
        (proc)
        (slot-fset t 'reloading-time-left (.reloading-period t)))
      (slot-fset t 'reloading-time-left (- (.reloading-time-left t) period))))

(define-method (step! tower-id (t <tower>))
  (set! towers (aa-insert towers
                          tower-id
                          (step t))))

(define-class <basic-turret> (<tower>)
  (firing :init-value #f #:accessor firing)
  (reloading-time-left #:init-value 2000 #:accessor .reloading-time-left)
  (reloading-period #:init-value 2000 #:accessor .reloading-period)
  (range #:init-value 92 #:accessor range)
  (shot-damage #:init-value 1 #:accessor shot-damage))

(define-method (step (t <basic-turret>))
  (catch 'no-candidate
    ;; Enemy in range
    (lambda ()
      (let ((nearest-enemy-id
             (nearest t enemies (range t))))
        (if (<= (.reloading-time-left t) period)
            (begin
              (damage! nearest-enemy-id (shot-damage t))
              (clone t
                     ((.reloading-time-left) (.reloading-period t))
                     ((firing) #t)))
            (clone t
                   ((.reloading-time-left) (- (.reloading-time-left t) period))
                   ((firing) #t)))))
    ;; Enemy not in range
    (lambda (_)
      (clone t
             ((.reloading-time-left) (- (.reloading-time-left t) period))
             ((firing) #f)))))

(define-method (draw (t <basic-turret>) position)
  (if (firing t)
      (draw (get-sprite
             (let ((r (random 30)))
               (cond ((= r 1) #:basic-turret-fire-0)
                     ((= r 2) #:basic-turret-fire-1)
                     ((= r 3) #:basic-turret-fire-2)
                     (else #:basic-turret)))) position)
      (draw (get-sprite #:basic-turret) position)))

(define-class <basic-turret-l2> (<basic-turret>)
  (reloading-time-left #:init-value 1429 #:accessor .reloading-time-left)
  (reloading-period #:init-value 1429 #:accessor .reloading-period)
  (range #:init-value 158 #:accessor range))

(define-method (draw (t <basic-turret-l2>) position)
  (if (firing t)
      (draw (get-sprite
             (let ((r (random 40)))
               (cond ((= r 1) #:basic-turret-l2-fire-0)
                     ((= r 2) #:basic-turret-l2-fire-1)
                     ((= r 3) #:basic-turret-l2-fire-2)
                     ((= r 4) #:basic-turret-l2-fire-3)
                     ((= r 5) #:basic-turret-l2-fire-4)
                     (else #:basic-turret-l2)))) position)
      (draw (get-sprite #:basic-turret-l2) position)))

(define-class <basic-turret-l3> (<basic-turret-l2>)
  (reloading-time-left #:init-value 952 #:accessor .reloading-time-left)
  (reloading-period #:init-value 952 #:accessor .reloading-period)
  (range #:init-value 190 #:accessor range))

(define-method (draw (t <basic-turret-l3>) position)
  (if (firing t)
      (draw (get-sprite
             (let ((r (random 50)))
               (cond ((= r 1) #:basic-turret-l3-fire-0)
                     ((= r 2) #:basic-turret-l3-fire-1)
                     ((= r 3) #:basic-turret-l3-fire-2)
                     ((= r 4) #:basic-turret-l3-fire-3)
                     ((= r 5) #:basic-turret-l3-fire-4)
                     ((= r 6) #:basic-turret-l3-fire-5)
                     ((= r 7) #:basic-turret-l3-fire-6)
                     ((= r 8) #:basic-turret-l3-fire-7)
                     (else #:basic-turret-l3)))) position)
      (draw (get-sprite #:basic-turret-l3) position)))

(define-class <delay-turret> (<tower>)
  (reloading-time-left #:init-value 2000 #:accessor .reloading-time-left)
  (reloading-period #:init-value 2000 #:accessor .reloading-period)
  (delay-duration #:init-value 6000 #:accessor delay-duration)
  (delay-factor #:init-value 0.5 #:accessor delay-factor)
  (range #:init-value 92 #:accessor range))

(define-method (draw (bt <delay-turret>) position)
  (draw (get-sprite #:delay-turret) position))

(define-method (step (t <delay-turret>))
  (step-reloading-time t (lambda ()
                           (catch 'no-candidate
                             (lambda ()
                               (let ((nearest-enemy-id
                                      (nearest t enemies (range t))))
                                 (affect-speed! nearest-enemy-id
                                                (aa-search enemies
                                                           nearest-enemy-id)
                                                (delay-factor t)
                                                (delay-duration t))))
                             (lambda (_) #nil)))))

(define-class <delay-turret-l2> (<delay-turret>)
  (reloading-time-left #:init-value 1429 #:accessor .reloading-time-left)
  (reloading-period #:init-value 1429 #:accessor .reloading-period)
  (delay-duration #:init-value 7000 #:accessor delay-duration)
  (delay-factor #:init-value 0.4 #:accessor delay-factor)
  (range #:init-value 158 #:accessor range))

(define-method (draw (bt <delay-turret-l2>) position)
  (draw (get-sprite #:delay-turret-l2) position))

(define-class <delay-turret-l3> (<delay-turret-l2>)
  (reloading-time-left #:init-value 952 #:accessor .reloading-time-left)
  (reloading-period #:init-value 952 #:accessor .reloading-period)
  (delay-duration #:init-value 8000 #:accessor delay-duration)
  (delay-factor #:init-value 0.3 #:accessor delay-factor))

(define-method (draw (bt <delay-turret-l3>) position)
  (draw (get-sprite #:delay-turret-l3) position))

(define-class <resource-tower> (<tower>)
  (reloading-time-left #:init-value 3000 #:accessor .reloading-time-left)
  (reloading-period #:init-value 3000 #:accessor .reloading-period))

(define-method (draw (rt <resource-tower>) position)
  (draw (get-sprite #:resource-tower) position))

(define-method (step (t <resource-tower>))
  (step-reloading-time t (lambda ()
                           (deposit-credits! 1))))

(define-class <resource-tower-l2> (<resource-tower>)
  (reloading-time-left #:init-value 2143 #:accessor .reloading-time-left)
  (reloading-period #:init-value 2143 #:accessor .reloading-period))

(define-method (draw (rt <resource-tower-l2>) position)
  (draw (get-sprite #:resource-tower-l2) position))

(define-class <resource-tower-l3> (<resource-tower-l2>)
  (reloading-time-left #:init-value 1429 #:accessor .reloading-time-left)
  (reloading-period #:init-value 1429 #:accessor .reloading-period))

(define-method (draw (rt <resource-tower-l3>) position)
  (draw (get-sprite #:resource-tower-l3) position))

(add-hook! update-hook
           (lambda (dt)
             (unless (game-paused?)
               (aa-for-each
                (lambda (kv)
                  (step! (car kv) (cdr kv)))
                towers))))
